# Display class

import matplotlib.pyplot as plt
import numpy as np
import math
import json


class Display:
    def __init__(self, listMachine, expe):
        """
        The Display class

        :param listMachine: List containing the machine objects (number, static power, states and name).
        :param expe: Experience object (powerStart, powerEnd et byStep).
        """
        self.listMachine = listMachine
        self.expe = expe
        self.x = list(range(self.expe.powerStart, self.expe.powerEnd, self.expe.byStep))
        self.minPower = self.setMinPower()
        self.maxPower = self.setMaxPower()
        self.numberMachine = self.setNumberMachine()
        self.display = self.dataMilp()
        self.displayH = self.dataHeuristics()
        self.displayRC = self.dataRandomChoice()
        self.error = self.dataError()
        self.errorBis = self.dataErrorBis()

    def setMinPower(self):
        minPower = math.inf
        for k in range(len(self.listMachine)):
            static = self.listMachine[k].static
            if static <= minPower:
                minPower = static - 1
        return minPower

    def setMaxPower(self):
        maxPower = 0
        for k in range(len(self.listMachine)):
            maxDynPower = self.listMachine[k].states[-1][0] * self.listMachine[k].states[-1][1] + \
                          self.listMachine[k].static
            if maxDynPower >= maxPower:
                maxPower = maxDynPower + 1
        return maxPower

    def setNumberMachine(self):
        numberMachine = 0
        for k in range(len(self.listMachine)):
            numberMachine += self.listMachine[k].number
        return numberMachine

    def dataMilp(self):
        nom_fichier_solution = "Solution_Milp_with_" + str(len(self.listMachine)) + "_types_and_" + \
                               str(self.numberMachine) + "_machine_from_" + \
                               str(self.expe.powerStart) + "_to_" + str(self.expe.powerEnd) + "_W_by_step_" + \
                               str(self.expe.byStep)
        with open(nom_fichier_solution, 'r') as file:
            display = json.load(file)
        return display

    def dataHeuristics(self):
        nom_fichier_solution = "Solution_Heuristics_with_" + str(len(self.listMachine)) + "_types_and_" + \
                               str(self.numberMachine) + "_machine_from_" + \
                               str(self.expe.powerStart) + "_to_" + str(self.expe.powerEnd) + "_W_by_step_" + \
                               str(self.expe.byStep)
        with open(nom_fichier_solution, 'r') as file:
            displayH = json.load(file)
        return displayH

    def dataRandomChoice(self):
        nom_fichier_solution = "Solution_RandomChoice_with_" + str(len(self.listMachine)) + "_types_and_" + \
                               str(self.numberMachine) + "_machine_from_" + \
                               str(self.expe.powerStart) + "_to_" + str(self.expe.powerEnd) + "_W_by_step_" + \
                               str(self.expe.byStep)
        with open(nom_fichier_solution, 'r') as file:
            displayRC = json.load(file)
        return displayRC

    def dataError(self):
        error = dict()
        error['RC'] = [x - y for x, y in zip(self.display['objectifMilp'],
                                                  self.displayRC['Random choice Average'])]
        error['BPP'] = [x - y for x, y in zip(self.display['objectifMilp'], self.displayH['Balance power perf'])]
        error['BSRWS'] = [x - y for x, y in zip(self.display['objectifMilp'], self.displayH['Best state rws'])]
        error['BSRWSAR'] = [x - y for x, y in zip(self.display['objectifMilp'], self.displayH['Best state rwsar'])]
        return error

    def dataErrorBis(self):
        errorBis = dict()
        errorBis['BPP'] = [x - y for x, y in
                                zip(self.display['objectifMilp'], self.displayH['Balance power perf'])]
        errorBis['BSRWS'] = [x - y for x, y in zip(self.display['objectifMilp'], self.displayH['Best state rws'])]
        errorBis['BSRWSAR'] = [x - y for x, y in
                                    zip(self.display['objectifMilp'], self.displayH['Best state rwsar'])]
        return errorBis

    def printMilp(self):
        print("----------------------------------- MILP :")
        print("Average Runtime : ", np.average(self.display['Time Milp']))
        print("Min Runtime : ", np.min(self.display['Time Milp']))
        print("Max Runtime : ", np.max(self.display['Time Milp']))

    def printRC(self):
        pourcent = []
        for x, y in zip(self.display['objectifMilp'], self.displayRC['Random choice Average']):
            pourcent.append((x - y) / x)
        print("----------------------------------- RC :")
        print("Average % : ", np.average(pourcent) * 100)
        print("Average : ", np.average(self.error['RC']))
        print("Median % : ", np.median(pourcent) * 100)
        print("Median : ", np.median(self.error['RC']))
        print("Min % : ", np.min(pourcent) * 100)
        print("Min : ", np.min(self.error['RC']))
        print("Max % : ", np.max(pourcent) * 100)
        print("Max : ", np.max(self.error['RC']))
        print("Std % : ", np.std(pourcent) * 100)
        print("Std : ", np.std(self.error['RC']))
        print("Var % : ", np.var(pourcent) * 100)
        print("Var : ", np.var(self.error['RC']))
        print("Average Runtime : ", np.average(self.displayRC['Time random choice']))
        print("Min Runtime : ", np.min(self.displayRC['Time random choice']))
        print("Max Runtime : ", np.max(self.displayRC['Time random choice']))

    def printBPP(self):
        pourcent = []
        for x, y in zip(self.display['objectifMilp'], self.displayH['Balance power perf']):
            pourcent.append((x - y) / x)
        print("----------------------------------- BPP :")
        print("Average % : ", np.average(pourcent) * 100)
        print("Average : ", np.average(self.error['BPP']))
        print("Median % : ", np.median(pourcent) * 100)
        print("Median : ", np.median(self.error['BPP']))
        print("Min % : ", np.min(pourcent) * 100)
        print("Min : ", np.min(self.error['BPP']))
        print("Max % : ", np.max(pourcent) * 100)
        print("Max : ", np.max(self.error['BPP']))
        print("Std % : ", np.std(pourcent) * 100)
        print("Std : ", np.std(self.error['BPP']))
        print("Var % : ", np.var(pourcent) * 100)
        print("Var : ", np.var(self.error['BPP']))
        print("Average Runtime : ", np.average(self.displayH['Time balance power perf']))
        print("Min Runtime : ", np.min(self.displayH['Time balance power perf']))
        print("Max Runtime : ", np.max(self.displayH['Time balance power perf']))

    def printBSRWS(self):
        pourcent = []
        for x, y in zip(self.display['objectifMilp'], self.displayH['Best state rws']):
            pourcent.append((x - y) / x)
        print("----------------------------------- BSRWS :")
        print("Average % : ", np.average(pourcent) * 100)
        print("Average : ", np.average(self.error['BSRWS']))
        print("Median % : ", np.median(pourcent) * 100)
        print("Median : ", np.median(self.error['BSRWS']))
        print("Min % : ", np.min(pourcent) * 100)
        print("Min : ", np.min(self.error['BSRWS']))
        print("Max % : ", np.max(pourcent) * 100)
        print("Max : ", np.max(self.error['BSRWS']))
        print("Std % : ", np.std(pourcent) * 100)
        print("Std : ", np.std(self.error['BSRWS']))
        print("Var % : ", np.var(pourcent) * 100)
        print("Var : ", np.var(self.error['BSRWS']))
        print("Average Runtime : ", np.average(self.displayH['Time best state rws']))
        print("Min Runtime : ", np.min(self.displayH['Time best state rws']))
        print("Max Runtime : ", np.max(self.displayH['Time best state rws']))

    def printBSRWSAR(self):
        pourcent = []
        for x, y in zip(self.display['objectifMilp'], self.displayH['Best state rwsar']):
            pourcent.append((x - y) / x)
        print("----------------------------------- BSRWSAR :")
        print("Average % : ", np.average(pourcent) * 100)
        print("Average : ", np.average(self.error['BSRWSAR']))
        print("Median % : ", np.median(pourcent) * 100)
        print("Median : ", np.median(self.error['BSRWSAR']))
        print("Min % : ", np.min(pourcent) * 100)
        print("Min : ", np.min(self.error['BSRWSAR']))
        print("Max % : ", np.max(pourcent) * 100)
        print("Max : ", np.max(self.error['BSRWSAR']))
        print("Std % : ", np.std(pourcent) * 100)
        print("Std : ", np.std(self.error['BSRWSAR']))
        print("Var % : ", np.var(pourcent) * 100)
        print("Var : ", np.var(self.error['BSRWSAR']))
        print("Average Runtime : ", np.average(self.displayH['Time best state rwsar']))
        print("Min Runtime : ", np.min(self.displayH['Time best state rwsar']))
        print("Max Runtime : ", np.max(self.displayH['Time best state rwsar']))

    def firstResults(self):
        self.printMilp()
        self.printRC()
        self.printBPP()
        self.printBSRWS()
        self.printBSRWSAR()

    def ratioWithStaticFunction(self):
        """
        Function that computes the ratio of each DVFS state of each machine depending on the power.
        """
        ratiosMachine = []
        for k in range(len(self.listMachine)):
            for stateNumberMachine in range(1, len(self.listMachine[k].states)):
                ratio = []
                for x in np.arange(self.minPower, self.maxPower, 0.1):
                    if x <= self.listMachine[k].static:
                        ratio.append(math.inf)
                    else:
                        dynamic = min(x - self.listMachine[k].static,
                                      self.listMachine[k].states[stateNumberMachine][0] *
                                      self.listMachine[k].states[stateNumberMachine][1])
                        ratio.append(x / (dynamic / self.listMachine[k].states[stateNumberMachine][1]))
                ratiosMachine.append([k, ratio, stateNumberMachine])
        return ratiosMachine

    def bestRatio(self, ratiosMachine):
        """
        Function that returns the best ratio of each type of machine whatever their DVFS state, depending on the power.

        :param ratiosMachine: List of ratios for each machine and for each DVFS state depending on the power.
        """
        bestRatioMachine = []
        for k in range(len(self.listMachine)):
            bestRatioMachineK = []
            bestStateMachineK = []
            for i in range(len(ratiosMachine[0][1])):
                currentBestRatio = math.inf
                currentBestState = 0
                for listRatio in ratiosMachine:
                    if listRatio[0] == k:
                        if listRatio[1][i] < currentBestRatio:
                            currentBestRatio = listRatio[1][i]
                            currentBestState = listRatio[2]
                bestStateMachineK.append(currentBestState)
                bestRatioMachineK.append(currentBestRatio)
            bestRatioMachine.append([bestRatioMachineK, bestStateMachineK])
        return bestRatioMachine

    def displayRatioMachines(self):
        """
        Function that displays the best ratio for each type of machine whatever the DVFS state depending on the power.
        """
        ratio = self.ratioWithStaticFunction()
        bestRatio = self.bestRatio(ratio)
        fig, ax = plt.subplots()
        colorLine = ["black", "#bae4bc", "#7bccc4", "#43a2ca", "#0868ac", "black", "#bae4bc", "#7bccc4", "#43a2ca",
                     "#0868ac", "black", "#bae4bc", "#7bccc4", "#43a2ca", "#0868ac"]
        for machine in range(len(bestRatio)):
            xRatio = list(np.arange(self.minPower, self.maxPower, 0.1))
            xRatioPrime = []
            for value in xRatio:
                xRatioPrime.append(value)
            ratioXMachine = []
            ratioYMachine = []
            for i in range(len(xRatio)):
                if self.listMachine[machine].static <= xRatioPrime[i] <= self.listMachine[machine].static + \
                        self.listMachine[machine].states[-1][0] * self.listMachine[machine].states[-1][1]:
                    ratioXMachine.append(xRatioPrime[i])
                    ratioYMachine.append(bestRatio[machine][0][i])
            ax.step(ratioXMachine, ratioYMachine, linewidth=5, label=self.listMachine[machine].name,
                    color=colorLine[machine])
        ax.set(xlim=(self.minPower, self.maxPower), ylim=(0.1, 2))
        ax.yaxis.grid(True, linestyle='-', which='major', alpha=0.5)
        ax.xaxis.grid(True, linestyle='-', which='major', alpha=0.5)
        ax.set_ylabel('Best ratio (W/GFlops)', fontsize=36)
        ax.set_xlabel('Power (W)', fontsize=36)
        ax.set_title('Best ratio depending on power', y=1.05, fontsize=38)
        ax.tick_params(axis='both', which='major', labelsize=35)
        lines, labels = ax.get_legend_handles_labels()
        ax.legend(lines, labels, loc="lower right", fancybox=True, framealpha=0.5, fontsize=35)
        plt.show()

    def displayObjectifs(self):
        """
        Function that displays the maximum computing power of the machines computed by the MILP and the heuristics
        depending on the power.
        """
        fig, ax = plt.subplots()
        X = [item * 0.001 for item in self.x]

        objMilp = [item * 0.001 for item in self.display['objectifMilp']]
        ax.step(X, objMilp, linewidth=5, label='MILP', color='black')

        objRc = [item * 0.001 for item in self.displayRC['Random choice Average']]
        ax.step(X, objRc, linewidth=5, label='RC', color='#bae4bc', linestyle='-')

        objRcMin = [item * 0.001 for item in self.displayRC['Random choice Min']]
        ax.step(X, objRcMin, linewidth=1, label='RC min-max', color='red', linestyle='-')

        objRcMax = [item * 0.001 for item in self.displayRC['Random choice Max']]
        ax.step(X, objRcMax, linewidth=1, color='red', linestyle='-')

        objBpp = [item * 0.001 for item in self.displayH['Balance power perf']]
        ax.step(X, objBpp, linewidth=5, label='BPP', color='#7bccc4', linestyle='-')

        objBsrws = [item * 0.001 for item in self.displayH['Best state rws']]
        ax.step(X, objBsrws, linewidth=5, label='BSRWS', color="#43a2ca", linestyle='-')

        objBsrwsar = [item * 0.001 for item in self.displayH['Best state rwsar']]
        ax.step(X, objBsrwsar, linewidth=5, label='BSRWSAR', color="#0868ac", linestyle='-')

        ax.set(xlim=(self.expe.powerStart * 0.001, 1.02 * self.expe.powerEnd * 0.001),
               ylim=(0, 1.02 * float(max(self.display['objectifMilp']) * 0.001)))
        ax.yaxis.grid(True, linestyle='-', which='major', alpha=0.5)
        ax.xaxis.grid(True, linestyle='-', which='major', alpha=0.5)
        ax.legend(loc='upper left', fancybox=True, framealpha=0.5, fontsize=35)
        # ax.set_ylabel('Computing power (TFlops)', fontsize=36)
        ax.set_xlabel('Power (kW)', fontsize=36)
        # ax.set_title('Computing power depending on power', y=1.05, fontsize=38)
        plt.xticks(fontsize=35)
        plt.yticks(fontsize=35)
        plt.show()

    def displayBoxplotsErrors(self):
        """
        Function that displays for each heuristic the boxplot of the difference from optimal solution.
        """
        fig, ax = plt.subplots()

        errorTflops = self.error

        errorTflops['RC'] = [item * 0.001 for item in errorTflops['RC']]

        errorTflops['BPP'] = [item * 0.001 for item in errorTflops['BPP']]

        errorTflops['BSRWS'] = [item * 0.001 for item in errorTflops['BSRWS']]

        errorTflops['BSRWSAR'] = [item * 0.001 for item in errorTflops['BSRWSAR']]

        bp = ax.boxplot(errorTflops.values(), showmeans=True, sym='+',
                        meanprops=dict(markeredgecolor='black', markerfacecolor='firebrick'),
                        medianprops=dict(color='firebrick'))
        col = ['#bae4bc', '#7bccc4', '#43a2ca', '#0868ac']
        for element in ['boxes', 'whiskers', 'fliers', 'means', 'medians', 'caps']:
            i = 0
            j = 0
            for box in bp[element]:
                box.set(color=col[i], linewidth=5)
                if element == 'whiskers' or element == 'caps':
                    if j == 0:
                        j += 1
                    else:
                        i += 1
                        j = 0
                else:
                    i += 1

        ax.set_xticklabels(errorTflops.keys())
        ax.yaxis.grid(True, linestyle='-', which='major', alpha=0.5)
        ax.set_ylabel('Difference from optimal (TFlops)', fontsize=36)
        ax.set_title('Difference from optimal for each heuristics', y=1.05, fontsize=38)
        plt.xticks(fontsize=35)
        plt.yticks(fontsize=35)
        lst = [errorTflops['RC'], errorTflops['BPP'], errorTflops['BSRWS'], errorTflops['BSRWSAR']]
        y_max = 0
        for i in range(len(lst)):
            y_med = np.percentile(lst[i], 100)
            if y_med > y_max:
                y_max = y_med
            text = ' μ={:.2f}\n median={:.2f}\n σ={:.2f}'.format(
                np.mean(lst[i]),
                np.percentile(lst[i], 50),
                np.std(lst[i]))
            ax.annotate(text, xy=(i + 0.7, y_med * 1.04), fontsize=32)
        ax.set(ylim=(- y_max * 0.10, y_max * 1.3))
        plt.show()

    def displayBoxplotsErrorsWithoutRc(self):
        """
        Function that displays for each heuristic the boxplot (except RC) of the difference from optimal solution
        """
        fig, ax = plt.subplots()

        errorTflops = self.errorBis

        errorTflops['BPP'] = [item * 0.001 for item in errorTflops['BPP']]

        errorTflops['BSRWS'] = [item * 0.001 for item in errorTflops['BSRWS']]

        errorTflops['BSRWSAR'] = [item * 0.001 for item in errorTflops['BSRWSAR']]

        bp = ax.boxplot(errorTflops.values(), showmeans=True, sym='+',
                        meanprops=dict(markeredgecolor='black', markerfacecolor='firebrick'),
                        medianprops=dict(color='firebrick'))
        col = ['#7bccc4', '#43a2ca', '#0868ac']
        for element in ['boxes', 'whiskers', 'fliers', 'means', 'medians', 'caps']:
            i = 0
            j = 0
            for box in bp[element]:
                box.set(color=col[i], linewidth=5)
                if element == 'whiskers' or element == 'caps':
                    if j == 0:
                        j += 1
                    else:
                        i += 1
                        j = 0
                else:
                    i += 1

        ax.set_xticklabels(errorTflops.keys())
        ax.yaxis.grid(True, linestyle='-', which='major', alpha=0.5)
        ax.set_ylabel('Difference from optimal (TFlops)', fontsize=36)
        ax.set_title('Difference from optimal for each heuristics', y=1.05, fontsize=38)
        plt.xticks(fontsize=35)
        plt.yticks(fontsize=35)
        lst = [errorTflops['BPP'], errorTflops['BSRWS'], errorTflops['BSRWSAR']]
        y_max = 0
        for i in range(len(lst)):
            y_med = np.percentile(lst[i], 100)
            if y_med > y_max:
                y_max = y_med
            text = ' μ={:.2f}\n median={:.2f}\n σ={:.2f}'.format(
                np.mean(lst[i]),
                np.percentile(lst[i], 50),
                np.std(lst[i]))
            ax.annotate(text, xy=(i + 0.7, y_med * 1.04), fontsize=32)
        ax.set(ylim=(- y_max * 0.10, y_max * 1.3))
        plt.show()

    def heuristicsRuntime(self):
        """
        Function that displays the execution time of the heuristics depending on the power.
        """
        fig, ax = plt.subplots()

        X = [item * 0.001 for item in self.x]

        ax.step(X, self.display['Time Milp'], linewidth=1, label='Time MILP', color='black')
        ax.step(X, self.displayRC['Time random choice'], linewidth=5, label='Time RC', color='#bae4bc')
        ax.step(X, self.displayH['Time balance power perf'], linewidth=5, label='Time BPP', color='#7bccc4')
        ax.step(X, self.displayH['Time best state rws'], linewidth=5, label='Time BSRWS', color='#43a2ca')
        ax.step(X, self.displayH['Time best state rwsar'], linewidth=5, label='Time BSRWSAR', color='#0868ac')

        ax.xaxis.grid(True, linestyle='-', which='major', alpha=0.5)
        ax.yaxis.grid(True, linestyle='-', which='major', alpha=0.5)
        ax.legend(loc='lower right', fancybox=True, framealpha=0.5, fontsize=35)
        ax.set_ylabel('Runtime (s)', fontsize=36)
        ax.set_xlabel('Power (kW)', fontsize=36)
        # ax.set_title('Runtime of MILP and the heuristics depending on power', y=1.05, fontsize=38)
        plt.xticks(fontsize=35)
        plt.yticks(fontsize=35)
        plt.yscale('log')
        plt.show()
