import experienceClass
import machinesClass
import milpClass
import bppHeuristic
import bsrwsHeuristic
import bsrwsarHeuristic
import displayResultsClass
import randomChoiceHeuristic
import json
import machinesData

###########################################
# MACHINES IN OBJECT-ORIENTED PROGRAMMING #
###########################################

taurus = machinesClass.Machine(150, machinesData.staticPowerTaurus, machinesData.statesTaurus, "Taurus")
parasilo = machinesClass.Machine(169, machinesData.staticPowerParasilo, machinesData.statesParasilo, "Parasilo")
grisou = machinesClass.Machine(0, machinesData.staticPowerGrisou, machinesData.statesGrisou, "Grisou")
grimoire = machinesClass.Machine(134, machinesData.staticPowerGrimoire, machinesData.statesGrimoire, "Grimoire")
graoully = machinesClass.Machine(145, machinesData.staticPowerGraoully, machinesData.statesGraoully, "Graoully")
gros = machinesClass.Machine(195, machinesData.staticPowerGros, machinesData.statesGros, "Gros")
chifflet = machinesClass.Machine(89, machinesData.staticPowerChifflet, machinesData.statesChifflet, "Chifflet")
grele = machinesClass.Machine(106, machinesData.staticPowerGrele, machinesData.statesGrele, "Grele")
gemini = machinesClass.Machine(34, machinesData.staticPowerGemini, machinesData.statesGemini, "Gemini")
graphite = machinesClass.Machine(91, machinesData.staticPowerGraphite, machinesData.statesGraphite, "Graphite")
orion = machinesClass.Machine(128, machinesData.staticPowerOrion, machinesData.statesOrion, "Orion")
chetemi = machinesClass.Machine(0, machinesData.staticPowerChetemi, machinesData.statesChetemi, "Chetemi")

# The machines selected for the experiment
listMachine = [taurus, parasilo, grimoire, graoully, gros, chifflet, grele, gemini, graphite, orion]

########################################################################################################################
########################################################################################################################

###############################
# POWER VALUES FOR SIMULATION #
###############################

powerStart = 63
powerEnd = 267064
byStep = 100

expe = experienceClass.Experience(powerStart, powerEnd, byStep)

########################################################################################################################
########################################################################################################################

#################################################
# DO THE FILES WITH THIS SIMULATION DATA EXIST? #
#################################################

# If True, read the files without running the Milp and/or the heuristics.
# If False, run the Milp and/or the heuristics

dataExist = True

# Milp
fileExistMilp = dataExist

# RC heuristic
fileExistRC = dataExist

# Others heuristics (BPP, BSRWS, BSWSAR)
fileExistHeuristics = dataExist

########################################################################################################################
########################################################################################################################

###################################
# RUN THE MILP AND THE HEURISTICS #
###################################

print('----------------------------------- MILP :')
milp = milpClass.Milp(listMachine, expe, resultMilpExists=fileExistMilp)
milp.saveData()
timeMilp = milp.getRuntime()

print('----------------------------------- RC :')
randomChoice = randomChoiceHeuristic.RandomChoiceHeuristic(listMachine, expe, resultHeuristicExists=fileExistRC)
randomChoice.computeRC()
objectifRCAverage = randomChoice.getObjectiveAverage()
objectifRCMin = randomChoice.getObjectiveMin()
objectifRCMax = randomChoice.getObjectiveMax()
timeRC = randomChoice.getRuntime()
xTimeRCAverage = randomChoice.getXRuntimeAverage()

print('----------------------------------- BPP :')
bpp = bppHeuristic.BPP(listMachine, expe, resultHeuristicExists=fileExistHeuristics, step=0.05)
bpp.computeBalancePowerPerformanceHeuristic()
objectifBPP = bpp.getObjectifBpp()
timeBPP = bpp.getRuntime()
xTimeBPP = bpp.getXRuntime()

print('----------------------------------- BSRWS :')
bsrws = bsrwsHeuristic.BSRWS(listMachine, expe, resultHeuristicExists=fileExistHeuristics)
bsrws.computeBSRWS()
objectifBSRWS = bsrws.getObjective()
timeBSRWS = bsrws.getRuntime()
xTimeBSRWS = bsrws.getXRuntime()

print('----------------------------------- BSRWSAR :')
bsrwsar = bsrwsarHeuristic.BSRWSAndRemoving(listMachine, expe, resultHeuristicExists=fileExistHeuristics)
bsrwsar.computeBSRWSAR()
objectifBSRWSAR = bsrwsar.getObjective()
timeBSRWSAR = bsrwsar.getRuntime()
xTimeBSRWSAR = bsrwsar.getXRuntime()

########################################################################################################################
########################################################################################################################

##################################
# SAVE RESULTS OF THE HEURISTICS #
##################################

numberMachine = 0
for k in range(len(listMachine)):
    numberMachine += listMachine[k].number

if not fileExistHeuristics:
    ResultsHeuristics = dict()
    ResultsHeuristics['Balance power perf'] = objectifBPP
    ResultsHeuristics['Time balance power perf'] = xTimeBPP
    ResultsHeuristics['Best state rws'] = objectifBSRWS
    ResultsHeuristics['Time best state rws'] = xTimeBSRWS
    ResultsHeuristics['Best state rwsar'] = objectifBSRWSAR
    ResultsHeuristics['Time best state rwsar'] = xTimeBSRWSAR
    nom_fichier_solution = "Solution_Heuristics_with_" + str(len(listMachine)) + "_types_and_" + \
                           str(numberMachine) + "_machine_from_" + str(powerStart) + "_to_" + \
                           str(powerEnd) + "_W_by_step_" + str(byStep)
    with open(nom_fichier_solution, 'w', encoding='utf-8') as f:
        json.dump(ResultsHeuristics, f, indent=4)

if not fileExistRC:
    ResultsHeuristics = dict()
    ResultsHeuristics['Random choice Average'] = objectifRCAverage
    ResultsHeuristics['Random choice Min'] = objectifRCMin
    ResultsHeuristics['Random choice Max'] = objectifRCMax
    ResultsHeuristics['Time random choice'] = xTimeRCAverage
    nom_fichier_solution = "Solution_RandomChoice_with_" + str(len(listMachine)) + "_types_and_" + \
                           str(numberMachine) + "_machine_from_" + str(powerStart) + "_to_" + \
                           str(powerEnd) + "_W_by_step_" + str(byStep)
    with open(nom_fichier_solution, 'w', encoding='utf-8') as f:
        json.dump(ResultsHeuristics, f, indent=4)

########################################################################################################################
########################################################################################################################

###################
# DISPLAY RESULTS #
###################

display = displayResultsClass.Display(listMachine, expe)
display.firstResults()
display.displayRatioMachines()
display.displayObjectifs()
display.displayBoxplotsErrors()
display.displayBoxplotsErrorsWithoutRc()
display.heuristicsRuntime()
